/*
Copyright (c) 2015  V.Ayllón.  All rights reserved.
 
 TwinTeethMC is distributed in the hope that it will be useful, but
 WITHOUT ANY WARRANTY.  No author or distributor accepts responsibility
 to anyone for the consequences of using it or for whether it serves any
 particular purpose or works at all, unless he says so in writing.  Refer
 to the GNU General Public License, Version 2 or later, for full details.
 
 Everyone is granted permission to copy, modify and redistribute
 TwinTeethMC, but only under the conditions described in the GNU General Public
 License.  A copy of this license is supposed to have been
 given to you along with TwinTeethMC so you can know your rights and
 responsibilities.  It should be in a file named COPYING.  Among other
 things, the copyright notice and this notice must be preserved on all
 copies.
 
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 */
import controlP5.*;

ControlP5 cp5;


String ourWeb = "http://www.diyouware.com"; 

boolean portStatus = false;  // false = no port
boolean connectStatus = false;  // false = disconected
boolean fileStatus = false;  // false = no loaded


// The tools ID's

int MAX_TOOL_NUMBER = 5;

int PEN_TOOL = 0;
int DRILL_TOOL = 1;
int LASER_TOOL = 2;
int EXTRUDER_TOOL = 3;
int DISPENSER_TOOL = 4;

int activeTool = PEN_TOOL;


String toolPropID = "TwinTeethTool";

int portNum = 0;

Button connectCtrl, loadFileCtrl, printCtrl, pauseCtrl, resetCtrl, heaterSetCtrl, heaterOffCtrl, heaterCheckCtrl, extrudeCtrl, extrudeReverseCtrl, focusCtrl, focusOneCtrl, dispenseCtrl, dispenseReverseCtrl;
Button  sendCtrl;
RadioButton laserDiodeCtrl, laserModeCtrl;
Numberbox laserFeedRateCtrl, laserPowerCtrl, heaterTempCtrl, extrudeLenCtrl, extrudeFeedRateCtrl, dispenseLenCtrl, dispenseFeedRateCtrl, focusOffsetCtrl;
ScrollableList portCtrl;
CheckBox invCtrl, peckCtrl, laserOnCtrl;
Textlabel tempLabelCtrl, extrudeLabelCtrl, dispenseLabelCtrl, offsetLabelCtrl;
Numberbox xOffsetCtrl, yOffsetCtrl;

Textarea echoWindow, msgWindow;
Textfield cmdLine;

String commandValue = "";

String[][] stepItems = new String[5][4];

PImage[] imgs = new PImage[20];


int ECHO_WINDOW_WIDTH = 250;
int CMD_WINDOW_WIDTH = ECHO_WINDOW_WIDTH-40;
int CMD_WINDOW_LENGTH = 20;
int MSG_WINDOW_LENGTH = 20;


int TAB_WIDTH = 100;
int TAB_HEIGHT = 25;

int DEF_LASER_FEED_RATE = 1200;
int defLaserFeedrate = DEF_LASER_FEED_RATE;

int DEF_LASER_POWER = 56;
int defLaserPower = DEF_LASER_POWER;

int DEF_HEATER_TEMP = 190;
int defHeaterTemp = DEF_HEATER_TEMP;

float DEF_FOCUS_OFFSET = 3.5;

String actualHeaterTemp = new String(" T: 0.0/0.0 Cº");

/**
 ** GUI initialization
 **/
void GUIinit() {

  // Load images for the buttons


  imgs[0] = loadImage("data/dremel.jpg");
  imgs[1] = loadImage("data/pentool.jpg");
  imgs[2] = loadImage("data/lasertool.jpg");
  imgs[3] = loadImage("data/extruder.jpg");
  imgs[4] = loadImage("data/dispensertool.jpg");


  // By default all controllers are stored inside Tab 'default' 
  // Add more tabs one for each tool

  cp5.addTab("tab1")
    .hide()
    .setLabel("Dremel")
    .setWidth(TAB_WIDTH)
    .setHeight(TAB_HEIGHT)
    .setId(DRILL_TOOL)
    .setColorBackground(GRAY)
    .setColorLabel(WHITE)
    .setColorActive(GREEN)
    .getCaptionLabel().setFont(mainFont)
    ;
  cp5.getTab("tab1").getCaptionLabel().toUpperCase(false);

  cp5.addTab("tab2")
    .hide()
    .setLabel("Laser")
    .setWidth(TAB_WIDTH)
    .setHeight(TAB_HEIGHT)
    .setId(LASER_TOOL)
    .setColorBackground(GRAY)
    .setColorLabel(WHITE)
    .setColorActive(GREEN)
    .getCaptionLabel().setFont(mainFont)
    ;
  cp5.getTab("tab2").getCaptionLabel().toUpperCase(false);

  cp5.addTab("tab3")
    .hide()
    .setLabel("3D printer")
    .setWidth(TAB_WIDTH)
    .setHeight(TAB_HEIGHT)
    .setId(EXTRUDER_TOOL)
    .setColorBackground(GRAY)
    .setColorLabel(WHITE)
    .setColorActive(GREEN)
    .getCaptionLabel().setFont(mainFont)
    ;

  cp5.getTab("tab3").getCaptionLabel().toUpperCase(false);

  cp5.addTab("tab4")
    .hide()
    .setLabel("Paste Dispenser")
    .setWidth(TAB_WIDTH)
    .setHeight(TAB_HEIGHT)
    .setId(DISPENSER_TOOL)
    .setColorBackground(GRAY)
    .setColorLabel(WHITE)
    .setColorActive(GREEN)
    .getCaptionLabel().setFont(mainFont)
    ;

  cp5.getTab("tab4").getCaptionLabel().toUpperCase(false);

  // Get the tabs and change some properties

  cp5.getTab("default")
    .setLabel("Pen tool")
    .setWidth(TAB_WIDTH)
    .setHeight(TAB_HEIGHT)
    .activateEvent(true)
    .setId(PEN_TOOL)
    .setColorBackground(GRAY)
    .setColorLabel(WHITE)
    .setColorActive(GREEN)
    .getCaptionLabel().setFont(mainFont)
    ;
  cp5.getTab("default").getCaptionLabel().toUpperCase(false);

  // The tool images and logo

  cp5.addButton("Diyouware")
    .setPosition(X_MAX_LENGTH, 4)
    .setImages(loadImage("data/diyouware0.jpg"), loadImage("data/diyouware1.jpg"), loadImage("data/diyouware2.jpg"))
    .updateSize()
    .moveTo("global")
    .unlock();

  cp5.addButton("pentool")
    .setPosition(X_TOOL_IMAGE, Y_TOOL_IMAGE)
    .setImage(imgs[1])
    .moveTo("default");


  cp5.addButton("dremel")
    .setPosition(X_TOOL_IMAGE, Y_TOOL_IMAGE)
    .setImage(imgs[0])
    .moveTo("tab1");


  cp5.addButton("lasertool")
    .setPosition(X_TOOL_IMAGE, Y_TOOL_IMAGE)
    .setImage(imgs[2])
    .moveTo("tab2");  


  cp5.addButton("extruderTool")
    .setPosition(X_TOOL_IMAGE, Y_TOOL_IMAGE)
    .setImage(imgs[3])
    .moveTo("tab3");  

  cp5.addButton("pasteDispenserTool")
    .setPosition(X_TOOL_IMAGE, Y_TOOL_IMAGE)
    .setImage(imgs[4])
    .moveTo("tab4");  


  // The ports list box

  portCtrl = cp5.addScrollableList("Ports")
    .setPosition(X_MARGIN, Y_MARGIN)
    .setSize(100, 60)
    .setItemHeight(20)
    .setBarHeight(15)
    .setColorBackground(BLUE)
    .setColorActive(RED)
    ;

  portCtrl.moveTo("global");
  cp5.getProperties().move(portCtrl, "default", "Nothing");

  // The connect button

  connectCtrl = cp5.addButton("Connect")
    .setLabel("Connect")
    .setPosition(X_MARGIN+105, Y_MARGIN)
    .setSize(100, 20)
    .setColorBackground(BLUE)
    .setColorActive(GREEN)
    .setBroadcast(true)
    .unlock();


  connectCtrl.moveTo("global");
  connectCtrl.getCaptionLabel().setFont(mainFont);
  connectCtrl.getCaptionLabel().toUpperCase(false);

  cp5.getProperties().move(connectCtrl, "default", "Nothing");
  updatePortList();

  // The load file button

  loadFileCtrl = cp5.addButton("LoadFile")
    .setLabel("Load File")
    .setPosition(X_MARGIN+105*2, Y_MARGIN)
    .setSize(100, 20)
    .setColorBackground(BLUE)
    .setColorActive(GREEN)
    .setBroadcast(true)
    .lock();

  loadFileCtrl.moveTo("global");
  loadFileCtrl.getCaptionLabel().setFont(mainFont);
  loadFileCtrl.getCaptionLabel().toUpperCase(false);
  cp5.getProperties().move(loadFileCtrl, "default", "Nothing");

  // The print button


  printCtrl = cp5.addButton("StartPrint")
    .setLabel("Run")
    .setPosition(X_MARGIN+105*3, Y_MARGIN)
    .setSize(100, 20)
    .setColorBackground(BLUE)
    .setColorActive(GREEN)
    .setBroadcast(true)
    .lock();
  ;  

  printCtrl.moveTo("global");
  printCtrl.getCaptionLabel().setFont(mainFont);
  printCtrl.getCaptionLabel().toUpperCase(false);
  cp5.getProperties().move(printCtrl, "default", "Nothing");
  // The pause button

  pauseCtrl = cp5.addButton("Pause")
    .setLabel("Pause")
    .setPosition(X_MARGIN+105*4, Y_MARGIN)
    .setSize(100, 20)
    .setColorBackground(BLUE)
    .setColorActive(GREEN)
    .setBroadcast(true)
    .lock();
  ;
  pauseCtrl.moveTo("global");
  pauseCtrl.getCaptionLabel().setFont(mainFont);
  pauseCtrl.getCaptionLabel().toUpperCase(false);
  cp5.getProperties().move(pauseCtrl, "default", "Nothing");

  // The Emergency stop button

  resetCtrl = cp5.addButton("EmergencyStop")
    .setImages(loadImage("data/stop0.jpg"), loadImage("data/stop1.jpg"), loadImage("data/stop2.jpg"))
    .setPosition(X_MAX_LENGTH/2+5, Y_MAX_LENGTH/2-90)
    .setBroadcast(true)
    .updateSize()
    .lock()
    ;

  resetCtrl.moveTo("global");
  cp5.getProperties().move(resetCtrl, "default", "Nothing");


  // XOffset field

  xOffsetCtrl=cp5.addNumberbox("XOffset")
    .setPosition(X_MAX_LENGTH/2-65, Y_MAX_LENGTH/2+21)
    .setLabel("X:")
    .setSize(45, 14)
    .setRange(-35, +35)
    .setMultiplier(1) // set the sensitifity of the numberbox
    .setDirection(Controller.HORIZONTAL) // change the control direction to left/right
    .lock()
    ;
  xOffsetCtrl.setValue(0);
  xOffsetCtrl.moveTo("global");
  controlP5.Label l = xOffsetCtrl.getCaptionLabel();
  l.getStyle().marginTop = -19; //move upwards (relative to button size)
  l.getStyle().marginLeft = -20; //move to the right
  l.setFont(createFont("Arial", 9));
  l.toUpperCase(false);

  // yOffset field

  yOffsetCtrl=cp5.addNumberbox("YOffset")
    .setPosition(X_MAX_LENGTH/2-65, Y_MAX_LENGTH/2+37)
    .setLabel("Y:")
    .setSize(45, 14)
    .setRange(-40, +40)
    .setMultiplier(1) // set the sensitifity of the numberbox
    .setDirection(Controller.HORIZONTAL) // change the control direction to left/right
    .lock()
    ;
  yOffsetCtrl.setValue(0);
  yOffsetCtrl.moveTo("global");
  l = yOffsetCtrl.getCaptionLabel();
  l.getStyle().marginTop = -19; //move upwards (relative to button size)
  l.getStyle().marginLeft = -21; //move to the right
  l.setFont(createFont("Arial", 9));
  l.toUpperCase(false);

  // Offset label

  offsetLabelCtrl = cp5.addTextlabel("OffsetLabel")
    .setText("Offset Origin (mm)")
    .setPosition(X_MAX_LENGTH/2-90, Y_MAX_LENGTH/2+3)
    .setColorValue(WHITE)
    .setFont(createFont("Arial", 10))
    ;
  offsetLabelCtrl.moveTo("global");


  // add a callback listener to refresh the simulation screen
  // for each value change.
  CallbackListener offsetCallBack = new CallbackListener() {
    public void controlEvent(CallbackEvent c) {
      if (c.getAction()==ControlP5.ACTION_BROADCAST) {
        if (fileStatus) {
          sim.loadObject(fileRow);
          sim.refresh();
        }
      }
    }
  };


  xOffsetCtrl.addCallback(offsetCallBack);
  yOffsetCtrl.addCallback(offsetCallBack);

  ///// Drill SECTION  /////
  // The peck test  checkbox

  peckCtrl = cp5.addCheckBox("peckTest")
    .setPosition(X_MAX_LENGTH-110, Y_MAX_LENGTH/2-25)
    .setColorForeground(color(0))
    .setColorActive(color(RED))
    .setColorLabel(color(255))
    .setSize(10, 10)
    .addItem("peck test", 0)
    ;

  peckCtrl.moveTo("tab1");
  peckCtrl.getItem(0).setValue(0);
  peckCtrl.getItem(0).lock();
  peckCtrl.getItem(0).getCaptionLabel().setFont(createFont("Arial", 9));
  peckCtrl.getItem(0).getCaptionLabel().toUpperCase(false);

  ///// Laser SECTION  /////
  // Select laser diodes
  laserDiodeCtrl =cp5.addRadioButton("SetLaserDiode")
    .setPosition(X_MAX_LENGTH-110, Y_MAX_LENGTH/2-80)
    .setNoneSelectedAllowed(false)
    .setSize(10, 10)
    .setColorForeground(color(120))
    .setColorActive(RED)
    .setColorLabel(color(255))
    .setItemsPerRow(3)
    .setSpacingColumn(15)
    .addItem("B", 1)
    .addItem("R", 2)
    .addItem("IR", 3)

    ;

  laserDiodeCtrl.moveTo("tab2");
  laserDiodeCtrl.activate(0);



  // Laser on

  laserOnCtrl = cp5.addCheckBox("LaserOn")
    .setPosition(X_MAX_LENGTH-30, Y_MAX_LENGTH/2-80)
    .setColorForeground(color(0))
    .setColorActive(color(GREEN))
    .setColorLabel(color(255))
    .setSize(10, 10)
    .addItem("On", 0)
    ;

  laserOnCtrl.moveTo("tab2");
  laserOnCtrl.getItem(0).setValue(0);
  laserOnCtrl.getItem(0).unlock();
  laserOnCtrl.getItem(0).getCaptionLabel().setFont(createFont("Arial", 9));
  laserOnCtrl.getItem(0).getCaptionLabel().toUpperCase(false);


  // Focus

  focusCtrl = cp5.addButton("Focus")
    .setPosition(X_MAX_LENGTH-110, Y_MAX_LENGTH/2-60)
    .setImages(loadImage("data/focus0.png"), loadImage("data/focus1.png"), loadImage("data/focus2.png"))
    .setSize(20, 20)
    .lock()
    ;
  focusCtrl.moveTo("tab2");

  focusOneCtrl = cp5.addButton("FocusOne")
    .setPosition(X_MAX_LENGTH-79, Y_MAX_LENGTH/2-60)
    .setImages(loadImage("data/focusone0.png"), loadImage("data/focusone1.png"), loadImage("data/focusone2.png"))
    .setSize(20, 20)
    .lock()
    ;
  focusOneCtrl.moveTo("tab2");

  // Focus offset

  focusOffsetCtrl=cp5.addNumberbox("FocusOffset")
    .setPosition(X_MAX_LENGTH-46, Y_MAX_LENGTH/2-60)
    .setLabel("Offset")
    .setSize(40, 14)
    .setRange(.1, 5.0)
    .setMultiplier(.1) // set the sensitifity of the numberbox
    .setDirection(Controller.HORIZONTAL) // change the control direction to left/right
    .lock()
    ;
  focusOffsetCtrl.setValue(DEF_FOCUS_OFFSET);
  focusOffsetCtrl.moveTo("tab2");
  focusOffsetCtrl.getCaptionLabel().setFont(createFont("Arial", 9));
  focusOffsetCtrl.getCaptionLabel().toUpperCase(false);

  // The invert image checkbox

  invCtrl = cp5.addCheckBox("InvertImage")
    .setPosition(X_MAX_LENGTH-110, Y_MAX_LENGTH/2-25)
    .setColorForeground(color(0))
    .setColorActive(color(RED))
    .setColorLabel(color(255))
    .setSize(10, 10)
    .addItem("Inv.", 0)
    ;

  invCtrl.moveTo("tab2");
  invCtrl.getItem(0).setValue(0);
  invCtrl.getItem(0).lock();
  invCtrl.getItem(0).getCaptionLabel().setFont(createFont("Arial", 9));
  invCtrl.getItem(0).getCaptionLabel().toUpperCase(false);

  // The laser mode checkbox

  // add a callback listener to refresh the simulation screen
  // for each value change.
  CallbackListener laserModeCallBack = new CallbackListener() {
    public void controlEvent(CallbackEvent c) {
      if (c.getAction()==ControlP5.ACTION_CLICK) {
        if (fileStatus) {
          fileStatus = false;
          printCtrl.lock();
          yOffsetCtrl.lock();
          xOffsetCtrl.lock();
          printState = IDLE_STATE;
          autoState = PRINTER_READY_STATE;
          printCtrl.setLabel("Run");
          sim.initObjectPosition();
          sim.clear();
          sim.cleanDrawWindow();
        }
      }
    }
  };

  laserModeCtrl = cp5.addRadioButton("LaserMode")
    .setPosition(X_MAX_LENGTH-75, Y_MAX_LENGTH/2-25)
    .setColorForeground(color(0))
    .setColorActive(color(RED))
    .setColorLabel(color(255))
    . setItemsPerRow(3) 
    .setSize(10, 10)
    .setSpacingColumn(25)
    .addItem("Plot", 0)
    .addItem("3D", 1)
    ;

  laserModeCtrl.moveTo("tab2");
  laserModeCtrl.getItem(0).setValue(0);
  laserModeCtrl.getItem(0).lock();
  laserModeCtrl.getItem(0).getCaptionLabel().setFont(createFont("Arial", 9));
  laserModeCtrl.getItem(0).getCaptionLabel().toUpperCase(false);
  // Add to callback
  laserModeCtrl.getItem(0).addCallback(laserModeCallBack);
  laserModeCtrl.getItem(1).addCallback(laserModeCallBack);

  // Laser power

  laserPowerCtrl=cp5.addNumberbox("LaserPower")
    .setPosition(X_MAX_LENGTH-110, Y_MAX_LENGTH/2-8)
    .setLabel("Laser Power")
    .setSize(100, 14)
    .setRange(10, 128)
    .setMultiplier(1) // set the sensitifity of the numberbox
    .setDecimalPrecision(0)
    .setDirection(Controller.HORIZONTAL) // change the control direction to left/right
    .lock()
    ;
  laserPowerCtrl.setValue(DEF_LASER_POWER);
  laserPowerCtrl.moveTo("tab2");
  laserPowerCtrl.getCaptionLabel().setFont(createFont("Arial", 9));
  laserPowerCtrl.getCaptionLabel().toUpperCase(false);

  // add a callback listener to refresh the simulation screen
  // for each value change.
  CallbackListener laserPowerCallBack = new CallbackListener() {
    public void controlEvent(CallbackEvent c) {
      if (c.getAction()==ControlP5.ACTION_RELEASE) {
         int p = int(laserPowerCtrl.getValue());
        if (p != defLaserPower) {
          setLaserPower(p);
          defLaserPower = p;
        }
      }
    }
  };

  laserPowerCtrl.addCallback(laserPowerCallBack);

  // Laser Feed Rate
  /**
   laserFeedRateCtrl=cp5.addNumberbox("LaserFeedRate")
   .setPosition(X_MAX_LENGTH-110, Y_MAX_LENGTH/2+24)
   .setLabel("Feed Rate (mm/min)")
   .setSize(100, 14)
   .setRange(200, 1200)
   .setMultiplier(1) // set the sensitifity of the numberbox
   .setDirection(Controller.HORIZONTAL) // change the control direction to left/right
   .lock()
   ;
   laserFeedRateCtrl.setValue(DEF_LASER_FEED_RATE);
   laserFeedRateCtrl.moveTo("tab2");
   laserFeedRateCtrl.getCaptionLabel().setFont(createFont("Arial", 9));
   laserFeedRateCtrl.getCaptionLabel().toUpperCase(false);
   **/

  ///// 3D PRINTER SECTION  /////
  int HEATER_CTRL_XPOS = X_MAX_LENGTH-140;
  int HEATER_CTRL_YPOS = Y_MAX_LENGTH/2+20;

  // Heater temperature

  heaterTempCtrl=cp5.addNumberbox("HeaterTemp")
    .setPosition(HEATER_CTRL_XPOS, HEATER_CTRL_YPOS)
    .setLabel("Heater")
    .setSize(30, 14)
    .setRange(0, 230)
    .setMultiplier(1) // set the sensitifity of the numberbox
    .setDirection(Controller.HORIZONTAL) // change the control direction to left/right
    .lock()
    ;
  heaterTempCtrl.setValue(DEF_HEATER_TEMP);
  heaterTempCtrl.moveTo("tab3");
  heaterTempCtrl.getCaptionLabel().setFont(createFont("Arial", 12));
  heaterTempCtrl.getCaptionLabel().toUpperCase(false);

  // Heater Set

  heaterSetCtrl = cp5.addButton("HeaterSet")
    .setPosition(HEATER_CTRL_XPOS+33, HEATER_CTRL_YPOS)
    .setLabel("Set")
    .setSize(20, 14) 
    .setColorBackground(BLUE)
    .lock()
    ;
  heaterSetCtrl.moveTo("tab3");
  heaterSetCtrl.getCaptionLabel().setFont(createFont("Arial", 9));
  heaterSetCtrl.getCaptionLabel().toUpperCase(false);

  // Heater off

  heaterOffCtrl = cp5.addButton("HeaterOff")
    .setPosition(HEATER_CTRL_XPOS+33+23, HEATER_CTRL_YPOS)
    .setLabel("Off")
    .setSize(20, 14) 
    .setColorBackground(BLUE)
    .lock()
    ;
  heaterOffCtrl.moveTo("tab3");
  heaterOffCtrl.getCaptionLabel().setFont(createFont("Arial", 9));
  heaterOffCtrl.getCaptionLabel().toUpperCase(false);

  // Heater check

  heaterCheckCtrl = cp5.addButton("HeaterCheck")
    .setPosition(HEATER_CTRL_XPOS+33+23+23, HEATER_CTRL_YPOS)
    .setLabel("Check")
    .setSize(33, 14) 
    .setColorBackground(BLUE)
    .lock()
    ;
  heaterCheckCtrl.moveTo("tab3");
  heaterCheckCtrl.getCaptionLabel().setFont(createFont("Arial", 9));
  heaterCheckCtrl.getCaptionLabel().toUpperCase(false);


  // Temperature label

  tempLabelCtrl = cp5.addTextlabel("TempLabel")
    .setText(actualHeaterTemp)
    .setPosition(HEATER_CTRL_XPOS+34, HEATER_CTRL_YPOS+18)
    .setColorValue(BLUE)
    .setFont(createFont("Arial", 12))
    ;

  tempLabelCtrl.moveTo("tab3");

  int EXTRUDE_CTRL_XPOS = X_MAX_LENGTH-140;
  int EXTRUDE_CTRL_YPOS = Y_MAX_LENGTH/2-20;

  // Extrude label

  extrudeLabelCtrl = cp5.addTextlabel("ExtrudeLabel")
    .setText("Extruder")
    .setPosition(EXTRUDE_CTRL_XPOS-3, EXTRUDE_CTRL_YPOS-20)
    .setColorValue(WHITE)
    .setFont(createFont("Arial", 12))
    ;
  extrudeLabelCtrl.moveTo("tab3");

  // Extrude reverse

  extrudeReverseCtrl = cp5.addButton("ExtrudeReverse")
    .setPosition(EXTRUDE_CTRL_XPOS, EXTRUDE_CTRL_YPOS)
    .setImages(loadImage("Arrow-Up0.png"), loadImage("Arrow-Up1.png"), loadImage("Arrow-Up2.png") )
    .setSize(25, 14) 
    .setColorBackground(BLUE)
    .lock()
    ;
  extrudeReverseCtrl.moveTo("tab3");
  extrudeReverseCtrl.getCaptionLabel().setFont(createFont("Arial", 9));
  extrudeReverseCtrl.getCaptionLabel().toUpperCase(false);

  // Extrude len

  extrudeLenCtrl=cp5.addNumberbox("ExtrudeLen")
    .setPosition(EXTRUDE_CTRL_XPOS+28, EXTRUDE_CTRL_YPOS)
    .setLabel("mm")
    .setSize(33, 14)
    .setRange(0, 999)
    .setMultiplier(1) // set the sensitifity of the numberbox
    .setDirection(Controller.HORIZONTAL) // change the control direction to left/right
    .lock()
    ;
  extrudeLenCtrl.setValue(0);
  extrudeLenCtrl.moveTo("tab3");
  extrudeLenCtrl.getCaptionLabel().setFont(createFont("Arial", 9));
  extrudeLenCtrl.getCaptionLabel().toUpperCase(false);

  // Extrude Feedrate

  extrudeFeedRateCtrl=cp5.addNumberbox("ExtrudeFeedRate")
    .setPosition(EXTRUDE_CTRL_XPOS+31+33, EXTRUDE_CTRL_YPOS)
    .setLabel("mm/min")
    .setSize(35, 14)
    .setRange(0, 500)
    .setMultiplier(1) // set the sensitifity of the numberbox
    .setDirection(Controller.HORIZONTAL) // change the control direction to left/right
    .lock()
    ;
  extrudeFeedRateCtrl.setValue(200);
  extrudeFeedRateCtrl.moveTo("tab3");
  extrudeFeedRateCtrl.getCaptionLabel().setFont(createFont("Arial", 9));
  extrudeFeedRateCtrl.getCaptionLabel().toUpperCase(false);

  // Extrude Forward

  extrudeCtrl = cp5.addButton("Extrude")
    .setPosition(EXTRUDE_CTRL_XPOS+31+33+38, EXTRUDE_CTRL_YPOS)
    .setImages(loadImage("Arrow-Down0.png"), loadImage("Arrow-Down1.png"), loadImage("Arrow-Down2.png") )
    .setSize(25, 14) 
    .setColorBackground(BLUE)
    .lock()
    ;
  extrudeCtrl.moveTo("tab3");
  extrudeCtrl.getCaptionLabel().setFont(createFont("Arial", 9));
  extrudeCtrl.getCaptionLabel().toUpperCase(false);
  //
  // PASTE DISPENSER
  //
  // Dispense label

  dispenseLabelCtrl = cp5.addTextlabel("DispenseLabel")
    .setText("Dispenser")
    .setPosition(EXTRUDE_CTRL_XPOS-3, EXTRUDE_CTRL_YPOS-20)
    .setColorValue(WHITE)
    .setFont(createFont("Arial", 12))
    ;
  dispenseLabelCtrl.moveTo("tab4");

  // Dispense reverse

  dispenseReverseCtrl = cp5.addButton("DispenseReverse")
    .setPosition(EXTRUDE_CTRL_XPOS, EXTRUDE_CTRL_YPOS)
    .setImages(loadImage("Arrow-Up0.png"), loadImage("Arrow-Up1.png"), loadImage("Arrow-Up2.png") )
    .setSize(25, 14) 
    .setColorBackground(BLUE)
    .lock()
    ;
  dispenseReverseCtrl.moveTo("tab4");
  dispenseReverseCtrl.getCaptionLabel().setFont(createFont("Arial", 9));
  dispenseReverseCtrl.getCaptionLabel().toUpperCase(false);

  // Dispense len

  dispenseLenCtrl=cp5.addNumberbox("DispenseLen")
    .setPosition(EXTRUDE_CTRL_XPOS+28, EXTRUDE_CTRL_YPOS)
    .setLabel("mm")
    .setSize(33, 14)
    .setRange(0.0, 10.0)
    .setMultiplier(.1) // set the sensitifity of the numberbox
    .setDirection(Controller.HORIZONTAL) // change the control direction to left/right
    .lock()
    ;
  dispenseLenCtrl.setValue(0.0);
  dispenseLenCtrl.moveTo("tab4");
  dispenseLenCtrl.getCaptionLabel().setFont(createFont("Arial", 9));
  dispenseLenCtrl.getCaptionLabel().toUpperCase(false);

  // Dispense Feedrate

  dispenseFeedRateCtrl=cp5.addNumberbox("DispenseFeedRate")
    .setPosition(EXTRUDE_CTRL_XPOS+31+33, EXTRUDE_CTRL_YPOS)
    .setLabel("mm/min")
    .setSize(35, 14)
    .setRange(0, 500)
    .setMultiplier(1) // set the sensitifity of the numberbox
    .setDirection(Controller.HORIZONTAL) // change the control direction to left/right
    .lock()
    ;
  dispenseFeedRateCtrl.setValue(200);
  dispenseFeedRateCtrl.moveTo("tab4");
  dispenseFeedRateCtrl.getCaptionLabel().setFont(createFont("Arial", 9));
  dispenseFeedRateCtrl.getCaptionLabel().toUpperCase(false);

  // Dispense Forward

  dispenseCtrl = cp5.addButton("Dispense")
    .setPosition(EXTRUDE_CTRL_XPOS+31+33+38, EXTRUDE_CTRL_YPOS)
    .setImages(loadImage("Arrow-Down0.png"), loadImage("Arrow-Down1.png"), loadImage("Arrow-Down2.png") )
    .setSize(25, 14) 
    .setColorBackground(BLUE)
    .lock()
    ;
  dispenseCtrl.moveTo("tab4");
  dispenseCtrl.getCaptionLabel().setFont(createFont("Arial", 9));
  dispenseCtrl.getCaptionLabel().toUpperCase(false);


  // The Text areas

  echoWindow = cp5.addTextarea("Echo")
    .setPosition(X_MAX_LENGTH, Y_MARGIN+50)
    .setSize(ECHO_WINDOW_WIDTH, Y_MAX_LENGTH-50)
    .setFont(textFont)
    .setLineHeight(14)
    .setColor(BLACK)
    .setColorBackground(WHITE)
    .setColorForeground(color(255, 100))
    ;
  cp5.getProperties().move(echoWindow, "default", "Nothing");

  echoWindow.moveTo("global");

  msgWindow = cp5.addTextarea("Messages")
    .hideScrollbar() 
    .setPosition(X_MIN_POS, Y_MAX_LENGTH+Y_MARGIN+5)
    .setSize(X_MAX_POS-5, MSG_WINDOW_LENGTH)
    .setFont(textFont)
    .setLineHeight(14)
    .setColor(BLACK)
    .setColorBackground(WHITE)
    .setColorForeground(BLACK)
    ;
  msgWindow.moveTo("global");
  cp5.getProperties().move(msgWindow, "default", "Nothing");

  // Command Line and send button

  cmdLine = cp5.addTextfield("CommandLine")
    .setLabel("G-Code")
    .setPosition(X_MAX_LENGTH, Y_MAX_LENGTH+Y_MARGIN+5)
    .setSize(CMD_WINDOW_WIDTH, CMD_WINDOW_LENGTH)
    .setFont(textFont)
    .setMax(50)
    .setColor(WHITE)
    .setColorBackground(BLACK)
    .setColorForeground(WHITE)
    .setAutoClear(false)
    .lock();
  cmdLine.moveTo("global");
  cp5.getProperties().move(cmdLine, "default", "Nothing");

  sendCtrl = cp5.addButton("SendGcodeButton")
    .setLabel("Send")
    .setBroadcast(false)
    .setPosition(X_MAX_LENGTH+CMD_WINDOW_WIDTH+5, Y_MAX_LENGTH+Y_MARGIN+5)
    .setSize(35, 20)
    .setValue(3)
    .setBroadcast(true)
    .setColorBackground(BLUE)
    .setColorActive(GREEN)
    .lock()
    ;

  sendCtrl.moveTo("global");
  sendCtrl.getCaptionLabel().setFont(mainFont);
  sendCtrl.getCaptionLabel().toUpperCase(false);

  cp5.getProperties().move(sendCtrl, "default", "Nothing");

  // Init the Jog control

  JOGinit();

  // Load the environment of tools

  for (int i = 0; i < MAX_TOOL_NUMBER; i++) {
    loadToolEnvironment(i);
  }

  // Load the default

  loadToolEnvironment();
}

/**
 * Diyouware() - Open our web page
 **/
void Diyouware(int v) {
  link(ourWeb);
}

/**
 ** LaserOn() - Turn laser on/off
 **/
void LaserOn(float[] v) {
  if (printState == IDLE_STATE) {
    if (v[0] == 1)
      turnLaserOn();
    else
      turnToolOff();
  }
}
/**
 ** LaserPower() - Set laser power
 
 void LaserPower(int p) {
 if (!(laserModeCtrl.getState(0) || laserModeCtrl.getState(1))) { // If laser tool in rastering
 if (printState == IDLE_STATE && p != defLaserPower) {
 setLaserPower(p);
 }
 } else { // 2D and 3D mode
 setLaserPower(p);
 }
 defLaserPower = p;
 }
 **/
/**
 ** LaserFeedRate() - Set laser feed rate
 **/
void LaserFeedRate(int f) {
  defLaserFeedrate = f;
}
/**
 ** InvertImage() - Invert image
 **/
void InvertImage(float[] f) {
  if (fileStatus && !(laserModeCtrl.getState(0) || laserModeCtrl.getState(1))) { // If laser tool in rastering
    if ( raster != null) {
      raster.invertImage();
    }
  }
}
/**
 ** LaserMode() - change laser mode Invert, 2D,3D. Not used.
 **/
void LaserMode(int mode) {
}
/**
 ** HeaterTemp() - Set heater temperature
 **/
void HeaterTemp(int t) {
  defHeaterTemp = t;
}

/**
 ** HeaterSet() - Set temperature button
 **/
void HeaterSet(int v) {
  setTargetTemp((int)heaterTempCtrl.getValue());
  tempLabelCtrl.setColor(WHITE);
  readCurrentTemp();
}
/**
 ** HeaterOff() - Set off the heater
 **/
void HeaterOff(int v) {
  setTargetTemp(0);
  tempLabelCtrl.setColor(BLUE);
  readCurrentTemp();
}
/**
 ** HeaterCheck() - Check heater temp
 **/
void HeaterCheck(int v) {
  readCurrentTemp();
}
/**
 ** ExtrudeReverse() - Extude reverse
 **/
void ExtrudeReverse(int v) {
  extrudeFilament((int)extrudeLenCtrl.getValue()*-1, (int)extrudeFeedRateCtrl.getValue());
}
/**
 ** ExtrudeLen() - Extrude Len
 **/
void ExtrudeLen(int l) {
}
/**
 ** ExtrudeFeedRate() - Exrtude FeedRate
 **/
void ExtrudeFeedRate(int f) {
}
/**
 ** Extrude() - Extrude some filament
 **/
void Extrude(int v) {
  extrudeFilament((int)extrudeLenCtrl.getValue(), (int)extrudeFeedRateCtrl.getValue());
}
/**
 ** DispenseReverse() - Dispens reverse
 **/
void DispenseReverse(int v) {
  dispensePaste(dispenseLenCtrl.getValue()*-1, (int)dispenseFeedRateCtrl.getValue());
}
/**
 ** DispenseLen() - Dispense Len
 **/
void DispenseLen(float l) {
}
/**
 ** DispenseFeedRate() - Dispense FeedRate
 **/
void DispenseFeedRate(int f) {
}
/**
 ** Dispense() - Dispense some paste
 **/
void Dispense(int v) {
  dispensePaste(dispenseLenCtrl.getValue(), (int)dispenseFeedRateCtrl.getValue());
}
/**
 * Print a message in the message window
 **/
void printMsg(String s) {
  msgWindow.setText(s);
}
/**
 * Print a message in the message window
 **/
void printlnMsg(String s) {
  msgWindow.setText(s+"\n");
}
/**
 * Print something in the echo window
 **/
void printlnEcho(String s) {

  echoWindow.append(s);
  echoWindow.scrolled(200);

  if (logger && logFile != null)
    logFile.println(s);
}

/**
 ** loadtToolEnvironment()
 **/
void loadToolEnvironment() {
  cp5.loadProperties("TwinTeethDefault");
}

/**
 ** loadtToolEnvironment()
 **/
void loadToolEnvironment(int tool) {
  cp5.loadProperties(toolPropID+tool);
}

/**
 ** saveToolEnvironment()
 **/
void saveToolEnvironment(int tool) {
  cp5.saveProperties(toolPropID+tool, toolPropID+tool);
}

/**
 ** saveToolEnvironment()
 **/
void saveToolEnvironment() {
  cp5.saveProperties("TwinTeethDefault", "default");
}

/**
 ** Class to catch the close window event()
 **/
public class DisposeHandler
{
  DisposeHandler(PApplet pa)
  {
    pa.registerMethod("dispose", this);
  }

  public void dispose()
  {

    // Turn laser off
    turnToolOff();

    // Save the environment 
    for (int i = 0; i < MAX_TOOL_NUMBER; i++) {
      saveToolEnvironment(i);
    }

    saveToolEnvironment();

    sim = null;
    System.gc();
  }
}