/*
Copyright (c) 2015  V.Ayllón.  All rights reserved.
 
 TwinTeethMC is distributed in the hope that it will be useful, but
 WITHOUT ANY WARRANTY.  No author or distributor accepts responsibility
 to anyone for the consequences of using it or for whether it serves any
 particular purpose or works at all, unless he says so in writing.  Refer
 to the GNU General Public License, Version 2 or later, for full details.
 
 Everyone is granted permission to copy, modify and redistribute
 TwinTeethMC, but only under the conditions described in the GNU General Public
 License.  A copy of this license is supposed to have been
 given to you along with TwinTeethMC so you can know your rights and
 responsibilities.  It should be in a file named COPYING.  Among other
 things, the copyright notice and this notice must be preserved on all
 copies.
 
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 */
import controlP5.*;

PrintWriter logFile = null;

// Printing states

int IDLE_STATE     = 0;
int PRINTING_STATE = 1;
int RESUME_STATE   = 2;

boolean printerStatus = false; //false not ready
int printState = IDLE_STATE;  // 0: Stopped, 1: Printing 2: Paused

/**
 ** StartPrint() - Inkoved from print button
 **/
void StartPrint(int v) {
  if (fileStatus) { // If file opened
    if (v == 1) {
      switch(printState) {
      case 0:  // IDLE_STATE - Waiting printing and user press "Print" or "Restart"
        printlnMsg("Printing...");
        echoWindow.clear();
        printCtrl.setLabel("Stop");
        pauseCtrl.setLabel("Pause");
        pauseCtrl.unlock();
        lockControls();
        rowNum = 0;  // Reset buffer #

        if (activeTool == LASER_TOOL) {
          if (!(laserModeCtrl.getState(0) || laserModeCtrl.getState(1))) { //Rastering
            if (!raster.initRastering(imageFilename)) {
              printerTimeOut= RASTER_READY_TIMEOUT;
              printState = PRINTING_STATE;
              autoState = WAITING_RASTER_STATE;
            }
          } else if (laserModeCtrl.getState(1)) { //Plotting with laser in 3D
            sim.init();
            setLaserPower(int(laserPowerCtrl.getValue()));
            initLaserMode(LASER_MODE_ON_E, int(laserDiodeCtrl.getValue()));
            printState = PRINTING_STATE;
            autoState = SEND_BUFFER_STATE;
          } else {  //Plotting with laser in 2D 
            sim.init();
            setLaserPower(int(laserPowerCtrl.getValue()));
            initLaserMode(LASER_MODE_ON_Z, int(laserDiodeCtrl.getValue()));
            printState = PRINTING_STATE;
            autoState = SEND_BUFFER_STATE;
          }
        } else if (activeTool == DRILL_TOOL) {
          sim.init();
          initSpindle();
          printState = PRINTING_STATE;
          autoState = SEND_BUFFER_STATE;
        } else if (activeTool == EXTRUDER_TOOL) {
          sim.init();
          printState = PRINTING_STATE;
          autoState = SEND_BUFFER_STATE;
        } else {
          sim.init();
          printState = PRINTING_STATE;
          autoState = SEND_BUFFER_STATE;
        }
        break;
      case 1:   // PRINTING_STATE -  We were printing and user press "Stop"
        stopPrinting(true);
        printlnMsg("Printing stopped.");
        printlnEcho("-- Printing stopped --\n");
        break;
      case 2: // RESUME_STATE - We were paused and user press "Resume"
        if (activeTool == LASER_TOOL) {
          if (lastDiode >= 0)
            turnLaserOn();
        }
        printlnMsg("Resuming...");
        printState = PRINTING_STATE;
        autoState = SEND_BUFFER_STATE;
        printCtrl.setLabel("Stop");
        pauseCtrl.setLabel("Pause");
        pauseCtrl.unlock();
        break;
      }
    }
  }
}
/**
 ** stopPrinting() - Stop actions
 **/
void stopPrinting(boolean sendCmd) {
  if (myPort != null) {
    myPort.clear();
  }
  if (sendCmd)
    sendStopPrintingCommands();
  printCtrl.setLabel("Restart");
  pauseCtrl.lock();
  unlockControls();
  printState = IDLE_STATE;
  autoState = PRINTER_READY_STATE;
}
int lastDiode = 0;
/**
 ** Pause() - We were printing and user press "Pause"
 **/
void Pause(int v) {

  if (v==1) {
    if (printState==PRINTING_STATE) {
      if (activeTool == LASER_TOOL) {
        turnToolOff();
      }
      printCtrl.setLabel("Resume");
      pauseCtrl.setLabel("Paused");

      printState = RESUME_STATE;
      autoState = PRINTER_READY_STATE;
      printlnMsg("Printing paused.");
      //      pauseCtrl.lock(); // Produce a bug with new ControlIP5, I don't know why
    }
  }
}

/**
 ** initSpindle() - init the spindle
 **/
void initSpindle() {
  setTestMode((int)peckCtrl.getItem(0).getValue());
  sendCommand(M03);
}

/**
 ** lockControls() - Lock/hide controls
 **/
void lockControls() {
  if (activeTool != PEN_TOOL) {
    cp5.getTab("default")
      .activateEvent(false)
      .hide();
  }
  if (activeTool != DRILL_TOOL) {
    cp5.getTab("tab1")
      .activateEvent(false)
      .hide();
  }    
  if (activeTool != LASER_TOOL) {
    cp5.getTab("tab2")
      .activateEvent(false)
      .hide();
  }    
  if (activeTool != EXTRUDER_TOOL) {
    cp5.getTab("tab3")
      .activateEvent(false)
      .hide();
  }
  if (activeTool != DISPENSER_TOOL) {
    cp5.getTab("tab4")
      .activateEvent(false)
      .hide();
  }
  loadFileCtrl.lock();
  sendCtrl.lock();
  cmdLine.lock();
  extrudeCtrl.lock();
  extrudeReverseCtrl.lock();
  extrudeLenCtrl.lock();
  extrudeFeedRateCtrl.lock();
  dispenseCtrl.lock();
  dispenseReverseCtrl.lock();
  dispenseLenCtrl.lock();
  dispenseFeedRateCtrl.lock();
  laserOnCtrl.getItem(0).lock();
  laserDiodeCtrl.getItem(0).lock();
  laserDiodeCtrl.getItem(1).lock();
  laserDiodeCtrl.getItem(2).lock();
  focusOffsetCtrl.lock();
  xOffsetCtrl.lock();
  yOffsetCtrl.lock();
  focusCtrl.lock();
  focusOneCtrl.lock();
  invCtrl.getItem(0).lock();
  laserModeCtrl.getItem(0).lock();
  peckCtrl.getItem(0).lock();
  lockJog(true);
}
/**
 ** unlockControls()) - Lock/show controls
 **/
void unlockControls() {

  cp5.getTab("default")
    .activateEvent(true)
    .show();


  cp5.getTab("tab1")
    .activateEvent(true)
    .show();


  cp5.getTab("tab2")
    .activateEvent(true)
    .show();


  cp5.getTab("tab3")
    .activateEvent(true)
    .show();

  cp5.getTab("tab4")
    .activateEvent(true)
    .show();

  loadFileCtrl.unlock();

  sendCtrl.unlock();
  cmdLine.unlock();
  //  laserFeedRateCtrl.unlock();
  laserPowerCtrl.unlock();
  heaterTempCtrl.unlock();
  heaterSetCtrl.unlock();
  heaterOffCtrl.unlock();
  heaterCheckCtrl.unlock();
  extrudeCtrl.unlock();
  extrudeReverseCtrl.unlock();
  extrudeLenCtrl.unlock();
  extrudeFeedRateCtrl.unlock();
  dispenseCtrl.unlock();
  dispenseReverseCtrl.unlock();
  dispenseLenCtrl.unlock();
  dispenseFeedRateCtrl.unlock();
  laserOnCtrl.getItem(0).unlock();
  laserDiodeCtrl.getItem(0).unlock();
  laserDiodeCtrl.getItem(1).unlock();
  laserDiodeCtrl.getItem(2).unlock();
  focusOffsetCtrl.unlock();
  xOffsetCtrl.unlock();
  yOffsetCtrl.unlock();
  focusCtrl.unlock();
  focusOneCtrl.unlock();
  invCtrl.getItem(0).unlock();
  laserModeCtrl.getItem(0).unlock();
  peckCtrl.getItem(0).unlock();
  lockJog(false);
}