/*
Copyright (c) 2015  V.Ayllón.  All rights reserved.
 
 TwinTeethMC is distributed in the hope that it will be useful, but
 WITHOUT ANY WARRANTY.  No author or distributor accepts responsibility
 to anyone for the consequences of using it or for whether it serves any
 particular purpose or works at all, unless he says so in writing.  Refer
 to the GNU General Public License, Version 2 or later, for full details.
 
 Everyone is granted permission to copy, modify and redistribute
 TwinTeethMC, but only under the conditions described in the GNU General Public
 License.  A copy of this license is supposed to have been
 given to you along with TwinTeethMC so you can know your rights and
 responsibilities.  It should be in a file named COPYING.  Among other
 things, the copyright notice and this notice must be preserved on all
 copies.
 
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 */
import controlP5.*;
import processing.serial.*;

MySerial myPort;        // The serial port

public static final int DISCONNECTED_STATE = 0;
public static final int WAIT_PRINTER_STATE = 1;
public static final int PRINTER_READY_STATE = 2;
public static final int SEND_BUFFER_STATE = 3;
public static final int WAITING_CTS_STATE = 4;
public static final int WAITING_FILE_STATE = 5;
public static final int DRAW_FILE_STATE = 6;
public static final int WAITING_RASTER_STATE = 7;

public static final int PRINTER_READY_TIMEOUT = 2000;
public static final int RASTER_READY_TIMEOUT = 18000;

public static int printerTimeOut = PRINTER_READY_TIMEOUT;

public static int autoState = DISCONNECTED_STATE;        // Automata state

public static String MSG_ERROR = "Error:";
public static String MSG_STARTED = "Started";
public static String MSG_TEMPERATURE = "ok T:";
public static String MSG_OK = "ok";
public static String MSG_RASTERING = "Rastering";
/**
 ** runStateMachine() - Run the machine
 **/
public void runStateMachine() {
  String myString;

  checkPortStatus();
  //System.out.println(autoState);
  switch(autoState) {
  case 0:    //  DISCONNECTED_STATE - Waiting ..
    delay(3);
    break;
  case 1:    //  WAIT_PRINTER_STATE - Port connected wait printer start message
    if (printerTimeOut-- <= 0) {
      printlnMsg("Error: Imposible to connect");
      printlnEcho("Error: Imposible to connect\n");
      setControls(false);
      connectCtrl.unlock();
      if (myPort != null)
        myPort.stop(); // Close the port
      connectStatus = false;
      printerStatus = false;
      printerTimeOut = PRINTER_READY_TIMEOUT;
      autoState = DISCONNECTED_STATE;
    } else {
      if (connectStatus) {
        while (myPort.available () > 0) {
          myString = myPort.readStringUntil('\n');
          if (myString != null) {
            printlnEcho(myString);
            if (codeSeen(myString, MSG_STARTED)) {
              printlnMsg("Connected to printer on "+SerialPortList.getPortNames()[portNum]+" @ "+BaudRate);
              echoWindow.setText("");  // Clear echo window
              printlnEcho("Connected\n");
              restartPrinter();
              break;
            }
          }
        }
      }
    }
    break;
  case 2: //  PRINTER_READY_STATE - Printer ready
    if (connectStatus) {
      while (myPort.available () > 0) {
        myString = myPort.readStringUntil('\n');
        if (myString != null) {
          if (codeSeen(myString, MSG_TEMPERATURE)) {  // Extruder Temperature
            actualHeaterTemp = getTempValue(myString);
            tempLabelCtrl.setText(actualHeaterTemp);
          } else if (codeSeen(myString, MSG_STARTED)) { // Reseted
            printlnMsg("Printer reseted...");
            restartPrinter();
          } else if (!codeSeen(myString, MSG_OK))
            printlnEcho(myString);
        }
      }
    }
    break;
  case 3: //  SEND_BUFFER_STATE - Send next file buffer
    myColor = blueRay;
    if ((buffer = sendBuffer()) == null) {
      stopPrinting(true);
      printState = IDLE_STATE;
      autoState = PRINTER_READY_STATE;  // End of file (or file error)
    } else {
      autoState = WAITING_CTS_STATE;  // Waiting CTS ///////////////////////////////////////////////////////////////////////////////////
    }
    break;

  case 4:    //  WAITING_CTS_STATE - Waiting OK from printer
    while (myPort.available () > 0) {
      myString = myPort.readStringUntil('\n');
      if (myString != null) {
        if (codeSeen(myString, MSG_TEMPERATURE)) {  // Extruder Temperature
          actualHeaterTemp = getTempValue(myString);
          tempLabelCtrl.setText(actualHeaterTemp);
        } else if (codeSeen(myString, MSG_OK)) {
          autoState = DRAW_FILE_STATE;
        } else if (codeSeen(myString, MSG_STARTED)) { // Reseted
          printlnMsg("Printer reseted...");
          restartPrinter();
        } else if (codeSeen(myString, MSG_ERROR)) { // Error
          stopPrinting(false);
          printlnMsg("Printer Error");
          printlnEcho(myString);
        } else
          printlnEcho(myString);
      }
    }
    break;
  case 5:    // WAITING_FILE_STATE - User selected a file, waiting openning file
    if (loadPath != null) {  // User selected a file
      if (activeTool == LASER_TOOL && !(laserModeCtrl.getState(0) || laserModeCtrl.getState(1))) { // If laser tool in rastering
        if (openPostScriptFile()) {
          autoState = PRINTER_READY_STATE;  // Error opening file
        } else {
          rowNum = 0;
          printState = IDLE_STATE;
          autoState = DRAW_FILE_STATE;
        }
      } else {
        fileRow = null;
        if (openFile()) {
          autoState = PRINTER_READY_STATE;  // Error opening file
        } else {
          sim.init();
          sim.cleanDrawWindow();
          sim.loadObject(fileRow);
          rowNum = 0;
          printState = IDLE_STATE;
          autoState = DRAW_FILE_STATE;
        }
      }
    } else {
      delay(10);
      autoState = WAITING_FILE_STATE;  // Waiting here till user select or cancel
    }
    break;
  case 6:    // DRAW_FILE_STATE - Simulation state,  print the file on screen
    sim.refresh();
    if (printState==PRINTING_STATE)
      autoState = SEND_BUFFER_STATE;
    else
      autoState = PRINTER_READY_STATE;    
    break;
  case 7:     //  WAITING_RASTER_STATE - Waiting init rastering from printer
    if (printerTimeOut-- <= 0) {
      printlnMsg("Error: waiting raster initialization");
      stopPrinting(false);
    } else {
      while (myPort.available () > 0) {
        myString = myPort.readStringUntil('\n');
        if (myString != null) {
          if (codeSeen(myString, MSG_RASTERING)) {
            autoState = SEND_BUFFER_STATE;
          } else if (codeSeen(myString, MSG_STARTED)) { // Reseted
            printlnMsg("Printer reseted...");
            restartPrinter();
          } else if (codeSeen(myString, MSG_ERROR)) { // Init raster error
            printlnMsg("Error initializing rastering. Try again");
            printlnEcho(myString);
            stopPrinting(false);
          } else {
            printlnEcho(myString);
          }
        }
      }
    }
    break;
  default:
    break;
  }
}

/**
 ** checkPortStatus() - check if the port is already active
 **/
void checkPortStatus() {

  if (SerialPortList.getPortNames() != null && SerialPortList.getPortNames().length <= 0) {
    portStatus = fileStatus = connectStatus = printerStatus = false;
    loadPath = null;
    portCtrl.clear();
    portCtrl.hide();
    connectCtrl.lock();
    setControls(false);
    autoState = DISCONNECTED_STATE;
    //    printlnMsg("Printer disconnected");
    //    printlnEcho("Disconnected\n");
  } else {
    if (portStatus == false) {
      updatePortList();
      portCtrl.show();
      connectCtrl.unlock();
      portStatus = true;
    } else if (autoState != WAIT_PRINTER_STATE) {
      connectCtrl.unlock();
    }
  }
}

/**
 ** codeSeen()
 **/
boolean codeSeen(String s, String code)
{
  String[] m1 = match(s, code);
  return (m1 != null);
}


/**
 ** getIntValue()  
 **/
int getIntValue(String s, int begin, int len)
{
  char c;
  int n = 0;
  int inx = begin; 
  char[] numbuf =  new char[len+1];

  while (inx < len) {
    c = s.charAt(inx++);
    if ((c >= '0') && (c <= '9')) {
      numbuf[n++] = c;
    } else { // end of number
      break;
    }
  }
  numbuf[n] = '\0';
  return int(new String(numbuf));
}

/**
 ** getTempValue()  - Get temperature value string
 **/
String getTempValue(String s)
{
  String[] m1 = match(s, " T:[0-9]*..[0-9]* /[0-9]*..[0-9]* ");
  if (m1 != null)
    return m1[0]+"Cº";
  else
    return ("-Err-");
}

/**
 ** sendCommand (String cmd)
 **/
public void sendCommand(String cmd) {
  if (!connectStatus) 
    return;

  try {
    if (activeTool == LASER_TOOL && !(laserModeCtrl.getState(0) || laserModeCtrl.getState(1))) {
      myPort.sendBufferedCommand(cmd, 1500, 20);
    } else {
      myPort.sendCommand(cmd);
    }
  } 
  catch (RuntimeException e) {
    printlnEcho(e.getMessage());
  }
}

/**
 ** sendBuffer() - Read next line from file and send it to serial line
 **/
String sendBuffer() {
  String buffer;
  if (activeTool == LASER_TOOL && !(laserModeCtrl.getState(0) || laserModeCtrl.getState(1))) {
    if ((buffer = raster.imageToGcode()) != null) {
      myPort.sendBufferedCommand(buffer);
      return buffer;
    } else {
      printlnMsg("-- End of job --");
      printlnEcho("-- End of file --\n");
      return null;
    }
  } else {
    if (rowNum < fileRow.length) {
      buffer = fileRow[rowNum];
      // 12-11-2015 - VAF - Added preprocessing of g-code command because
      // last versions of Eagle light only allows positive coordinates.
      myPort.sendCommand(sim.preProcessCommand(buffer));
      rowNum++;
      return buffer;
    } else {
      // Stop reading because of an error or file is empty
      printlnMsg("-- End of job --");
      printlnEcho("-- End of file --\n");
      //  printCtrl.setValue(0);  // Stop button
      return null;
    }
  }
}

/**
 ** Connect() with the printer
 **/
public void Connect(int v) {
  if (portStatus) {
    if (v==1 && connectStatus) {
      stopPrinting(false);
      if (myPort != null) {
        myPort.clear();
        myPort.stop();
      }
      if (logger && logFile != null) {
        logFile.flush(); // Writes the remaining data to the file
        logFile.close(); // Finishes the file
      }
      sim.initObjectPosition();
      sim.clear();
      sim.cleanDrawWindow();
      setControls(false);
      connectStatus = fileStatus = printerStatus = false;
      loadPath = null;
      fileRow = null;
      if (activeTool == LASER_TOOL && !(laserModeCtrl.getState(0) || laserModeCtrl.getState(1)))
        raster.reset();
      activeTool = PEN_TOOL;
      printlnMsg("Printer disconnected");
      printlnEcho("Disconnected\n");
      autoState = DISCONNECTED_STATE;
    } else {
      try {
        myPort = new MySerial(SerialPortList.getPortNames()[portNum], BaudRate);
      } 
      catch (Exception e) {
        printlnMsg("Error trying to connect to serial port: "+SerialPortList.getPortNames()[portNum]);       // Error
        //        updatePortList();
        return;
      }
      connectCtrl.lock();
      portCtrl.hide();
      myPort.clear();
      //      setControls(true);
      if (logger)
        logFile = createWriter(tempDirectory+ "/TwinTeethMCLog.txt");
      printlnMsg("Waiting printer on "+SerialPortList.getPortNames()[portNum]+" @ "+BaudRate);
      cmdLine.clear();
      connectStatus = true;
      autoState = WAIT_PRINTER_STATE;
    }
  }
}


/**
 ** updatePortList()
 **/
void updatePortList() {
  portCtrl.clear();

  if (SerialPortList.getPortNames() != null && SerialPortList.getPortNames().length > 0) {
    for (int i=0; i < SerialPortList.getPortNames ().length; i++) {
      portCtrl.addItem(SerialPortList.getPortNames()[i], i);
    }
    portCtrl.setValue(0);
    portNum=0;
    portStatus = true;
  } else {
    portStatus = false;
  }
}

/**
 ** EmergencyStop()  - The stop button
 **/
void EmergencyStop(int v) {
  autoState = DISCONNECTED_STATE;
  resetCtrl.lock();
  connectCtrl.lock();
  printCtrl.lock();
  pauseCtrl.lock();
  lockControls();
  printlnMsg("EStop button pressed");
  printlnEcho("-- Printer reseted --\n");
  // Reopen port to reset Arduino
  try {
    if (myPort != null)
      myPort.stop(); // Close the port
    myPort = new MySerial(SerialPortList.getPortNames()[portNum], BaudRate);
  } 
  catch (Exception e) {
    printlnMsg("Error trying to connect to serial port: "+SerialPortList.getPortNames()[portNum]);       // Error
    return;
  }

  loadPath = null;
  fileStatus = printerStatus = false;
  pauseCtrl.setLabel("Pause");
  printCtrl.setLabel("Run");
  raster.reset();
  sim.initObjectPosition();
  sim.cleanDrawWindow();
  printState = IDLE_STATE;
  autoState = WAIT_PRINTER_STATE;
}

/**
 ** restartPrinter() - The printer was externally reseted
 **/
void restartPrinter() {
  delay(5);
  loadPath = null;
  fileStatus = false;
  printerStatus = true;
  sendCommand(G54+activeTool);
  setDefSeekRate((int)xyFeedCtrl.getValue());
  setZMin(zMinCtrl[activeTool].getValue());
  unlockControls();
  setControls(true);
  pauseCtrl.lock();
  pauseCtrl.setLabel("Pause");
  printCtrl.setLabel("Run");
  raster.reset();
  sim.initObjectPosition();
  sim.cleanDrawWindow();
  printState = IDLE_STATE;
  autoState = PRINTER_READY_STATE;
}


/**
 ** setControls() - set controls
 **/
void setControls(boolean status) {

  if (status) {
    cp5.getTab("default")
      .activateEvent(true)
      .show();


    cp5.getTab("tab1")
      .activateEvent(true)
      .show();


    cp5.getTab("tab2")
      .activateEvent(true)
      .show();


    cp5.getTab("tab3")
      .activateEvent(true)
      .show();

    cp5.getTab("tab4")
      .activateEvent(true)
      .show();
    ;
    portCtrl.hide();
    connectCtrl.setLabel("Disconnect");
    connectCtrl.unlock();
    loadFileCtrl.unlock();
    printCtrl.lock();
    pauseCtrl.lock();
    resetCtrl.unlock();
    sendCtrl.unlock();
    cmdLine.unlock();
    //   laserFeedRateCtrl.unlock();
    laserPowerCtrl.unlock();
    heaterTempCtrl.unlock();
    heaterSetCtrl.unlock();
    heaterOffCtrl.unlock();
    heaterCheckCtrl.unlock();
    extrudeCtrl.unlock();
    extrudeReverseCtrl.unlock();
    extrudeLenCtrl.unlock();
    extrudeFeedRateCtrl.unlock();
    dispenseCtrl.unlock();
    dispenseReverseCtrl.unlock();
    dispenseLenCtrl.unlock();
    dispenseFeedRateCtrl.unlock();
    laserOnCtrl.getItem(0).unlock();
    laserDiodeCtrl.getItem(0).unlock();
    laserDiodeCtrl.getItem(1).unlock();
    laserDiodeCtrl.getItem(2).unlock();
    focusOffsetCtrl.unlock();
    focusCtrl.unlock();
    focusOneCtrl.unlock();
    invCtrl.getItem(0).unlock();
    laserModeCtrl.getItem(0).unlock();
    peckCtrl.getItem(0).unlock();
    lockJog(false);
  } else {
    portCtrl.show();
    cp5.getTab("tab1")
      .activateEvent(false)
      .hide();
    ;
    cp5.getTab("tab2")
      .activateEvent(false)
      .hide();
    ;
    cp5.getTab("tab3")
      .activateEvent(false)
      .hide();
    ;
    cp5.getTab("tab4")
      .activateEvent(false)
      .hide();
    ;
    cp5.getTab("default").bringToFront();
    connectCtrl.setLabel("Connect");
    loadFileCtrl.lock();
    printCtrl.lock();
    pauseCtrl.lock();
    resetCtrl.lock();
    sendCtrl.lock();
    cmdLine.lock();
    //   laserFeedRateCtrl.lock();
    laserPowerCtrl.lock();
    heaterTempCtrl.lock();
    heaterSetCtrl.lock();
    heaterOffCtrl.lock();
    heaterCheckCtrl.lock();
    extrudeCtrl.lock();
    extrudeReverseCtrl.lock();
    extrudeLenCtrl.lock();
    extrudeFeedRateCtrl.lock();
    dispenseCtrl.lock();
    dispenseReverseCtrl.lock();
    dispenseLenCtrl.lock();
    dispenseFeedRateCtrl.lock();
    laserOnCtrl.getItem(0).lock();
    laserDiodeCtrl.getItem(0).lock();
    laserDiodeCtrl.getItem(1).lock();
    laserDiodeCtrl.getItem(2).lock();
    focusOffsetCtrl.lock();
    xOffsetCtrl.lock();
    yOffsetCtrl.lock();
    focusCtrl.lock();
    focusOneCtrl.lock();
    invCtrl.getItem(0).lock();
    laserModeCtrl.getItem(0).lock();
    peckCtrl.getItem(0).lock();
    lockJog(true);
  }
}