/*
Copyright (c) 2015  V.Ayllón.  All rights reserved.
 
 TwinTeethMC is distributed in the hope that it will be useful, but
 WITHOUT ANY WARRANTY.  No author or distributor accepts responsibility
 to anyone for the consequences of using it or for whether it serves any
 particular purpose or works at all, unless he says so in writing.  Refer
 to the GNU General Public License, Version 2 or later, for full details.
 
 Everyone is granted permission to copy, modify and redistribute
 TwinTeethMC, but only under the conditions described in the GNU General Public
 License.  A copy of this license is supposed to have been
 given to you along with TwinTeethMC so you can know your rights and
 responsibilities.  It should be in a file named COPYING.  Among other
 things, the copyright notice and this notice must be preserved on all
 copies.
 
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 */
import java.util.Locale;
import controlP5.*;

boolean logger = false;  // true activate the logger
int BaudRate = 115200;

public static final int BLUE_DIODE = 1;
public static final int RED_DIODE  = 2;
public static final int IR_DIODE = 3;

public static final int DEFAULT_FOCUS_DIODE = IR_DIODE;

public static final int LASER_MODE_OFF = 0;     // Laser mode deactivated
public static final int LASER_MODE_RASTER = 1;  // Laser in rastering mode
public static final int LASER_MODE_ON_E = 2;    // Laser is fired on E axis
public static final int LASER_MODE_ON_Z = 3;    // Laser is fired on Z axis <= 0

public static final boolean ON = true;
public static final boolean OFF = false;

DisposeHandler endHndl;

float x;
float y;

boolean mouseLocked = false;

int X_MAX_POS = 650;
int X_MIN_POS = 0;
int Y_MAX_POS = 650;
int Y_MIN_POS = 0;

int X_MAX_LENGTH = (X_MAX_POS - X_MIN_POS);
int Y_MAX_LENGTH = (Y_MAX_POS - Y_MIN_POS);

int XTRA_WIDTH = 350;  // Extra length for command window and controls
int XTRA_LENGTH = 100;  // Extra length for msg window and top gui

int WINDOW_WIDTH  = X_MAX_LENGTH+XTRA_WIDTH;
int WINDOW_LENGTH = Y_MAX_LENGTH+XTRA_LENGTH;

int X_MARGIN = 3;
int Y_MARGIN = 30;

// Machine limits

int X_MAX_PRT = 30;
int X_MIN_PRT = -30;
int Y_MAX_PRT = 40;
int Y_MIN_PRT = -40;
int Z_MAX_PRT = 155;
int Z_MIN_PRT = -4;

int BLACK = color(0, 0, 0);
int WHITE = color(255, 255, 255);
int BLUE = color(2, 52, 77);
int DARKBLUE =color(255, 128, 0);
int RED = color(242, 0, 0);
int IRRED = color(255, 119, 119);
int GREEN = color(0, 160, 100);
int PURPLE = color(100, 0, 255);
int GREY = color(128);
int G3DDBACKGROUND = color(131, 191, 194);

color blueRay = color(100, 0, 255);
color grayRay = color(184, 134, 11);
color myColor = grayRay;
color posColor = GREY;
color fixtureColor = RED;
color fixtureAlpha = 0;

int myColorBackground = GREY;

PFont mainFont;
PFont textFont;

String loadPath = null;
BufferedReader fileReader = null;
String buffer = null;
String fileRow[] = null;
int rowNum = 0;

int[] DRAW_WINDOW_COORDS = {
  X_MARGIN, 380, X_MAX_LENGTH-(X_MARGIN*2), Y_MAX_LENGTH+Y_MARGIN
};

float DEF_MATERIAL_WIDTH = 70*2; // Bottom layer need more room because Eagle Bug
float DEF_MATERIAL_HEIGHT = 80*2;
int  DEF_MATERIAL_MARGIN = 0;



final String tempDirectory = System.getenv("TEMP");

Simulation sim = null;
Raster raster = null;

HpglParser hpgl = null;

void setup() {
  
  size(910, 720, P3D); // We can´t not use variables here
  org.apache.log4j.BasicConfigurator.configure();  // Init log4j
  endHndl = new DisposeHandler(this);
  mainFont =   createFont("Arial Bold", 12);
  textFont =   createFont("Arial", 12);
  Locale.setDefault(Locale.US); 
  imageMode(CORNERS);
  noStroke();
  cp5 = new ControlP5(this);
  cp5.getProperties().setFormat(ControlP5.SERIALIZED);  // Serialized property file format
  background(0);
  autoState = DISCONNECTED_STATE;
  raster = new Raster();
  sim = new Simulation();
  hpgl = new HpglParser();
  GUIinit();

  background(myColorBackground);
  sim.cleanDrawWindow();
  // Create a new file in the sketch directory
} 

void draw() {
  smooth();
  imageMode(CORNERS);
  background(myColorBackground);
  sim.updateGraphics();
  runStateMachine();
}


float[] coord = new float[2];
/**
 ** controlEvent()
 **/
void controlEvent(ControlEvent theEvent) {
  float value;


  if (theEvent.isFrom(laserDiodeCtrl)) {
    // If blue diode selected change color
    int purple = int(theEvent.getGroup().getArrayValue()[0]);
    int red = int(theEvent.getGroup().getArrayValue()[1]);
    if (purple ==1)
      laserDiodeCtrl.setColorActive(PURPLE);
    else if (red ==1)
      laserDiodeCtrl.setColorActive(RED);
    else
      laserDiodeCtrl.setColorActive(IRRED);
  }
  if (theEvent.isFrom(portCtrl)) {
    portNum = (int)theEvent.getValue();
  }

  // Change Active tool when change tab

  if (theEvent.isTab()) {
    int tool = theEvent.getTab().getId();
    if (tool != activeTool) {
      changeActiveTool(tool);
    }
  }

  // Check if the event is from the jog axis

  if (theEvent.getId() >= X_JOG_GROUP) {
    coord = jogCtrl.getArrayValue();
    Map m = stepCtrl.getItem((int)stepCtrl.getValue());
    String s1 = (String)m.get("text");
    value = Float.parseFloat(s1);
    if (theEvent.getId() == X_JOG_GROUP) {
      if (theEvent.getValue() < 0)
        value = -value;
      coord[0] += value;
      jogCtrl.setArrayValue(coord);
      relativeMove(value, 0, 0, (int)xyFeedCtrl.getValue());
    } else  if (theEvent.getId() == Y_JOG_GROUP) {
      if (theEvent.getValue() < 0)
        value = -value;
      coord[1] += value;
      jogCtrl.setArrayValue(coord);
      relativeMove(0, value*-1, 0, (int)xyFeedCtrl.getValue());
    } else if (theEvent.getId() == Z_JOG_GROUP) {
      relativeMove(0, 0, theEvent.getValue(), (int)zFeedCtrl.getValue());
    }
  }
}

/**
 ** changeActiveTool() - actions to change tool
 **/
void changeActiveTool(int tool) {
  activeTool = tool;
  sendCommand(G54+tool);
  setZMin(zMinCtrl[tool].getValue());
  if (activeTool == LASER_TOOL) {
    raster.reset();
  }
  fileStatus = false;
  printCtrl.lock();
  yOffsetCtrl.lock();
  xOffsetCtrl.lock();
  printCtrl.setLabel("Run");
  sim.initObjectPosition();
  sim.cleanDrawWindow();
  printState = IDLE_STATE;
  autoState = PRINTER_READY_STATE;
}

/**
 **  mousePressed() -  Mouse control methods
 **/
void mousePressed(MouseEvent event) {

  if ((mouseX > DRAW_WINDOW_COORDS[0] && mouseX < DRAW_WINDOW_COORDS[2] && 
    mouseY >DRAW_WINDOW_COORDS[1] && mouseY < DRAW_WINDOW_COORDS[3]) && fileStatus == true && printState == IDLE_STATE) {
    if (event.getCount()==2) {
      if (fileStatus) {
        sim.initObjectPosition();
        autoState = DRAW_FILE_STATE;
      }
    } else {
      mouseLocked = true;
      sim.setOffsetPoint(mouseX, mouseY);
    }
  } else {
    mouseLocked = false;
  }
}
/**
 **  mouseDragged() - Mouse control methods
 **/
void mouseDragged() {

  if (mouseLocked) {
    if (fileStatus) {
      if (keyPressed)
        sim.rotateObject(mouseX, mouseY);
      else
        sim.moveObject(mouseX, mouseY);
      autoState = DRAW_FILE_STATE;
    }
  }
}
/**
 **  mouseReleased() - Mouse control methods
 **/
void mouseReleased() {
  mouseLocked = false;
}
/**
 **  Processing 3.0 mouseWheel() - Mouse control methods
 **/
void mouseWheel(MouseEvent event) {
  int delta = (int)event.getCount();
  if ((mouseX > DRAW_WINDOW_COORDS[0] && mouseX < DRAW_WINDOW_COORDS[2] && 
    mouseY >DRAW_WINDOW_COORDS[1] && mouseY < DRAW_WINDOW_COORDS[3]) && fileStatus == true && printState == IDLE_STATE) {
    sim.scaleObject(delta);
    autoState = DRAW_FILE_STATE;  // Redraw
  }
}